-------------------------------------------------------------------------------------------------
-- Main add-on object
-------------------------------------------------------------------------------------------------
GuildNoteSetter = LibStub("AceAddon-3.0"):GetAddon("GuildNoteSetter")
local L = LibStub("AceLocale-3.0"):GetLocale("GuildNoteSetter", true)

local blCommPrefix = "GBlackList";
local commPrefix = "GNoteSetter";
local commCmdVerReq = "PollVerReq";
local commCmdVerResp = "Ver|";
local commCmdBid = "Bid|";

local addonVer = 3;

------------------------------------------------------------------------------------------------
-- Utility functions
-------------------------------------------------------------------------------------------------
local startsWith = function(str, prefix)
	return string.sub(str, 1, string.len(prefix)) == prefix
end

local split = function(str, sep)
	local sep, fields = sep or ":", {}
	local pattern = string.format("([^%s]+)", sep)
	str:gsub(pattern, function(c) fields[#fields+1] = c end)
	return fields
end

-------------------------------------------------------------------------------------------------
-- Roster class
-------------------------------------------------------------------------------------------------
local Roster = GuildNoteSetter:new_class(
	function(obj)
		obj.players = {}
		obj.ranks = {}
	end)

function Roster:OnRosterUpdate()
	self.players = {}
	for i = 1, GetNumGuildMembers() do
		local name, _, rank = GetGuildRosterInfo(i)
		if name then
			self.players[name] = {};
			self.players[name].index = i;
			self.players[name].rank = rank + 1;
		end
	end
end

function Roster:OnRanksUpdate()
	self.ranks = self.ranks or {}
	for i = 1, GuildControlGetNumRanks() do
		self.ranks[i] = self.ranks[i] or {}
		self.ranks[i].name = GuildControlGetRankName(i);

		-- The following commented out due to Blizzard forbidding these interface functions in v7.3
--		GuildControlSetRank(i);
--		local flags = { GuildControlGetRankFlags() }
--		self.ranks[i].canWriteOfficerChat = flags[4];
--		self.ranks[i].canEditPubNote = flags[10];
	end
end

function Roster:IsEmpty()
	return next(self.players) == nil
end

function Roster:PlayerExists(name)
	if self.players[name] then
		return true
	else
		return false
	end
end

function Roster:PlayerExistsAndUpToDate(name)
	-- Ensure player exists in our data
	if ((not self.players[name]) or (not self.players[name].index)) then
		return false
	end

	-- Ensure the player index is still correct
	local name2, _ = GetGuildRosterInfo(self.players[name].index);
	return name == name2
end

-- PlayerExistsAndUpToDate() shall be checked before calling this
function Roster:SetPlayerPublicNote(name, note)
	GuildRosterSetPublicNote(self.players[name].index, note);
end

function Roster:CanPlayerEditPubNotes(name)
	local player = self.players[name];
	if ((not player) or (not player.rank)) then
		return nil;
	end

	local rankInfo = self.ranks[player.rank];
	if ((not rankInfo) or (not rankInfo.canEditPubNote)) then
		return nil;
	end

	return rankInfo.canEditPubNote;
end

function Roster:IsPlayerTrustedBLSrc(name)
	local player = self.players[name];
	if ((not player) or (not player.rank)) then
		return nil;
	end

	local rankInfo = self.ranks[player.rank];
	if ((not rankInfo) or (not rankInfo.blTrusted)) then
		return nil;
	end

	return rankInfo.blTrusted;
end

function Roster:CanPlayerEditBL(name)
	local player = self.players[name];
	if ((not player) or (not player.rank)) then
		return false;
	end

	local rankInfo = self.ranks[player.rank];
	if ((not rankInfo) or (not rankInfo.canEditBL)) then
		return nil;
	end

	return rankInfo.canEditBL;
end

-------------------------------------------------------------------------------------------------
-- GuildNoteSetter life cycle management
-------------------------------------------------------------------------------------------------
function GuildNoteSetter:OnInitialize()
	self.db = LibStub("AceDB-3.0"):New("GuildNoteSetterDB")

	self:RegisterChatCommand("gns", "SlashCmd")
	self:RegisterChatCommand("gnotesetter", "SlashCmd")
end

function GuildNoteSetter:OnEnable()
	self.pollingVers = false;
	self.passedInit = false;
	self.ourRealmName = string.gsub(GetRealmName(), "%s+", "")     -- Realm name with no spaces
	self.peerVers = {};
	self.roster = Roster();
	self.announcments = {};
	self.debug = false;
	self.bidsRxLog = {};
	self.pendingTimers = {};

	self.blOutdated = false;
	self.blLastAnnounceTime = 0
	self.blInTransferFrom = nil
	self.blSourcesForUpdate = {}
	self.blOutTransfers = 0
	self.blPeerVers = {};
	self.detectedPlayers = {}

	self:RegisterEvent("GUILD_ROSTER_UPDATE")
	self:RegisterEvent("PLAYER_GUILD_UPDATE")
	self:RegisterEvent("GUILD_RANKS_UPDATE")
	self:RegisterEvent("CHAT_MSG_SYSTEM")
	self:RegisterEvent("PLAYER_REGEN_ENABLED")

	self:RegisterComm(commPrefix, "OnCommReceived");
	self:RegisterComm(blCommPrefix, "OnBlackListCommReceived");

	GuildRoster();
end

function GuildNoteSetter:OnDisable()
	self:UnregisterEvent("CHAT_MSG_SYSTEM")
	if self.passedInit then
		self:UnregisterEvent("CHAT_MSG_GUILD")
		self:UnregisterEvent("CHAT_MSG_WHISPER")
	end
	self:UnregisterEvent("GUILD_RANKS_UPDATE")
	self:UnregisterEvent("GUILD_ROSTER_UPDATE")
	self:UnregisterEvent("PLAYER_GUILD_UPDATE")
	self:UnregisterEvent("PLAYER_REGEN_ENABLED")
end

-- This is postponed since GetGuildInfo() does not return valid data at the beginning
function GuildNoteSetter:LateInit()
	self.passedInit = true;

	self.guildName = GetGuildInfo("player");
	if self.guildName == nil then
		self.passedInit = false
		return
	end
	self.playerName = self:LongName(GetUnitName("player"))

	self:RegisterEvent("CHAT_MSG_WHISPER")
	self:RegisterEvent("CHAT_MSG_GUILD")

	self:CreateBlackListUI();
	self:ScheduleNonCombatTimer("AnnounceOurBlackListDb", 30)		-- First announcement after 10 sec
	self:ScheduleNonCombatTimer("ParseGuildInfo", 12)
end

-------------------------------------------------------------------------------------------------
-- GuildNoteSetter event handlers
-------------------------------------------------------------------------------------------------
function GuildNoteSetter:CHAT_MSG_WHISPER(eventName, msg, author, lang, status, mid, unk, lineID, senderGuid)
	if not self:ShouldHandleNotes() then return; end

	local newNote = self:IsNoteSetRequest(msg);
	if newNote == nil then return end

	author = self:LongName(author);
	self:UpdateNoteOrRetry(author, newNote, true);
end

function GuildNoteSetter:CHAT_MSG_GUILD(eventName, msg, author, lang, lineID, senderGuid)
	if not self:ShouldHandleNotes() then return; end

	local newNote = self:IsNoteSetRequest(msg);
	if not newNote then return end

	author = self:LongName(author);
	self:UpdateNoteOrRetry(author, newNote, false);
end

function GuildNoteSetter:CHAT_MSG_SYSTEM(eventName, msg)
	if (string.find(msg, L["has joined the guild"]) ~= nil) then
		GuildRoster();
	end
end

function GuildNoteSetter:GUILD_ROSTER_UPDATE(eventName)
	if not self.passedInit then
		self:LateInit();
	end 

	self.roster:OnRosterUpdate();
	self:ScheduleNonCombatTimer("DetectBlackListedPlayers", 5);
end

function GuildNoteSetter:GUILD_RANKS_UPDATE(eventName)
	self.roster:OnRanksUpdate()
end

function GuildNoteSetter:PLAYER_GUILD_UPDATE(eventName)
	self.guildName = GetGuildInfo("player");
	if self.guildName ~= nil then
		self:LateInit();
	end 
end

-------------------------------------------------------------------------------------------------
-- Slash commands
-------------------------------------------------------------------------------------------------
function GuildNoteSetter:SlashCmd(input)
	if input == "pollver" then
		self:PollVersions()
	elseif input == "pollblver" then
		self:BlacklistPollDbVersions()
	elseif input == "ranks" then
		self:PrintRankNumbers()
	else
		self:Printf(L["Version %s"], self:GetVersion());
	end
end

function GuildNoteSetter:PollVersions()
	self:Print(L["Polling guild mates, please wait..."]);

	self.peerVers = {};
	self.pollingVers = true;
	self:SendCommMessage(commPrefix, commCmdVerReq, "GUILD");

	local VERSION_POLLING_PERIOD_sec = 10
	self:ScheduleTimer(self.VerPollTimerEnded, VERSION_POLLING_PERIOD_sec, self);
end

function GuildNoteSetter:VerPollTimerEnded()
	self.pollingVers = false;

	local s = "";
	local names = {}
    for n in pairs(self.peerVers) do
		table.insert(names, n)
	end
    table.sort(names)
    for i, n in ipairs(names) do
		local canEdit = self.roster:CanPlayerEditPubNotes(n);
		if canEdit then s = s .. "*"; end
		s = s .. self:ShortName(n) .. "(" .. self.peerVers[n] .. "), "
	end

	self:Printf(L["Found %d guild mates with the addon installed:"], table.getn(names));
	self:Print(s);
end

function GuildNoteSetter:PrintRankNumbers()
	self:Print(L["Guild rank numbers:"]);
	for i = 1, GuildControlGetNumRanks() do
		self:Print(i .. " - " ..GuildControlGetRankName(i));
	end
end

-------------------------------------------------------------------------------------------------
-- Communication handling
-------------------------------------------------------------------------------------------------
function GuildNoteSetter:OnCommReceived(prefix, msg, distribution, sender)
	if prefix ~= commPrefix or distribution ~= "GUILD" then return end

	sender = self:LongName(sender);
	if msg == commCmdVerReq then
		self:SendCommMessage(commPrefix, commCmdVerResp .. self:GetVersion(), "GUILD");
		return
	elseif startsWith(msg, commCmdBid) then
		self:OnBidReceived(msg, sender);
	elseif self.pollingVers and startsWith(msg, commCmdVerResp) then
		local ver = string.sub(msg, string.len(commCmdVerResp) + 1);
		self.peerVers[sender] = ver;
	end
end

function GuildNoteSetter:OnBidReceived(msg, sender)
	-- Save bids for debug purposes if enabled
	if self.debug then
		table.insert(self.bidsRxLog, {GetTime(), sender, msg});
	end

	-- Ignore bids from ourselves
	if sender == self.playerName then return; end

	-- Check if sender can edit the notes at all
	if not self.roster:CanPlayerEditPubNotes(sender) then return; end

	local parts = split(string.sub(msg, string.len(commCmdBid) + 1), "|");
	if not parts or not parts[1] or not parts[2] then return end

	local player = self:LongName(parts[1])
	local bid = tonumber(parts[2])

	if ((self.announcments[player]) and (self.announcments[player].bid < bid)) then
		self.announcments[player] = nil;		-- We been overridden, don't announce
	end
end

-------------------------------------------------------------------------------------------------
-- "Out of combat" timer (if elapses when in combat, the callback is executed one combat is over)
-------------------------------------------------------------------------------------------------
function GuildNoteSetter:ScheduleNonCombatTimer(func, delay, ...)
	self:ScheduleTimer("NonCombatTimerHandler", delay, func, ...)
end

function GuildNoteSetter:NonCombatTimerHandler(func, ...)
	if InCombatLockdown() then
		local timer = {...};
		timer.func = func
		timer.argsCount = select("#", ...)

		table.insert(self.pendingTimers, timer);
	else
		self[func](self, ...);
	end
end

function GuildNoteSetter:PLAYER_REGEN_ENABLED()
	if InCombatLockdown() then return end

	-- We left combat state, execute pending timers
	local timer = self.pendingTimers[1]
	if timer then
		table.remove(self.pendingTimers, 1);

		self[timer.func](self, unpack(timer, 1, timer.argsCount));

		if self.pendingTimers[1] then		-- More timers left
			self:ScheduleTimer("PLAYER_REGEN_ENABLED", 0.1)
		end
	end
end

-------------------------------------------------------------------------------------------------
-- Guild info parsing (for obtaining configuration)
-------------------------------------------------------------------------------------------------
function GuildNoteSetter:ParseGuildInfo()
	local info = GetGuildInfoText()
	if not info then
		self:Debug("Failed obtaining guild info.");
		return;
	end

	local linesArray = { string.split("\n", info) }
	local pattern = "@GNS:([^;]*);([^;]*);([^;]*);?(.*)"

	for _, line in pairs(linesArray) do
		local publicNoteList, blTrustList, blEditList, altRank = line:match(pattern);
		if publicNoteList then		
			local canEditPubNotes = self:ParseNumList(publicNoteList or "");
			local blTrusted = self:ParseNumList(blTrustList or "");
			local canEditBL = self:ParseNumList(blEditList or "");
			self.roster.ranks = self.roster.ranks or {};
			
			for i = 1, GuildControlGetNumRanks() do
				self.roster.ranks[i] = self.roster.ranks[i] or {};
				self.roster.ranks[i].canEditBL = canEditBL[i] or false;
				self.roster.ranks[i].blTrusted = blTrusted[i] or false;
				self.roster.ranks[i].canEditPubNote = canEditPubNotes[i] or false;
			end			

			-- self.altRank = altRank
			-- self:Debug("Alt rank was set to", self.altRank, " (", GuildControlGetRankName(self.altRank), ")")
			return;
		end
	end

	self:Print(L["No configuration found. Please add line in the following format to guild info: @GNS:[ranks-that-can-edit-public-notes];[ranks-trusted-for-blacklist-download];[rank-that-can-edit-balcklist]"]);
	self:Print("");
	self:Print(L["For example: @GNS:1,3,4;1,2;1,2"]);
	self:Print(L["To see rank numbers use: /gns ranks"]);
end

-------------------------------------------------------------------------------------------------
-- Internal functions
-------------------------------------------------------------------------------------------------
function GuildNoteSetter:GetVersion()
	return GetAddOnMetadata(self:GetName(), "Version") or "unknown";
end

function GuildNoteSetter:Debug(...)
	if self.debug then
		self:Print(...);
	end
end

function GuildNoteSetter:ShouldHandleNotes()
	return IsInGuild() and CanEditPublicNote();
end

function GuildNoteSetter:LongName(name)
	if (string.find(name, "%-") ~= nil) then
		return name;
	end

	return name .. "-" .. self.ourRealmName;
end

function GuildNoteSetter:ShortName(name)
	local index = string.find(name, "%-");
	if not index then
		return name;
	end

	return name:sub(1, index - 1);
end

function GuildNoteSetter:IsNoteSetRequest(msg)
	if not startsWith(string.lower(msg), L["!note"]) then
		return nil;
	end

	local newNote = strtrim(string.sub(msg, string.len(L["!note"]) + 1));
	return newNote;
end

function GuildNoteSetter:UpdateNoteOrRetry(player, newNote, responsePrivate)
	if not player or player == "" then return; end

	if (not responsePrivate) then
		local myBid = fastrandom(1000) + addonVer * 1000 + 1100000;	-- Bid higher than all older versions, this effectively disables them
		self:SendCommMessage(commPrefix, commCmdBid .. player .. "|" .. myBid, "GUILD");
		self.announcments[player] = { bid = myBid, success = false };

		local DELAY_FOR_ANNOUNCEMENT_sec = 2.5
		self:ScheduleTimer(self.AnnounceNoteUpdate, DELAY_FOR_ANNOUNCEMENT_sec, self, player);
	end

	if (self.roster:PlayerExistsAndUpToDate(player)) then
		self:UpdateNote(player, newNote, responsePrivate);
	else
		self:Debug("Scheduling update for later");

		-- Refresh roster and try again
		GuildRoster();

		local DELAY_FOR_ROSTER_UPDATE_WAIT_sec = 2
		self:ScheduleTimer(self.UpdateNote, DELAY_FOR_ROSTER_UPDATE_WAIT_sec, self, player, newNote, responsePrivate);
	end
end

-- This updates note if possible now, if not it gives up
function GuildNoteSetter:UpdateNote(player, newNote, responsePrivate)
	self:Debug("UpdateNote");
	if (not self.roster:PlayerExistsAndUpToDate(player)) then
		self.announcments[player] = nil;
		self:Printf(L["Failed updating the note for %s"], player);
		return;
	end

	self.roster:SetPlayerPublicNote(player, newNote);
	if responsePrivate then
		SendChatMessage(L["Note updated."], "WHISPER", nil, player);
	else
		if self.announcments[player] then
			self.announcments[player].success = true;
		end
	end
end

function GuildNoteSetter:AnnounceNoteUpdate(player)
	self:Debug("AnnounceNoteUpdate");
	if self.announcments[player] and self.announcments[player].success then
		SendChatMessage(L["Note for X updated."](player), "GUILD");
	end

	self.announcments[player] = nil;
end

function GuildNoteSetter:ParseNumList(list)
	local parts = { string.split(",", list) }
	local res = {}

	for _, a in pairs(parts) do
		res[tonumber(a)] = true;
	end
	
	return res
end
